/* GWNUM -- IBDWT FFT package for large numbers.

  Copyright 1996-2005 George Woltman.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; see the file COPYING.  If not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/

/*----------------------------------------------------------------------
| gwutil.c
|
| This file contains various utility routines that may be used by gwnum
| routines, prime95, or PRP.
+---------------------------------------------------------------------*/

/* Include files */

#include <stdlib.h>
#ifdef _WIN64
#include <malloc.h>
#endif
#include "gwutil.h"


/* Aligned malloc routines.  MSVC 8 supports these in the C runtime library. */
/* Emulate these routines for other ports. */

void * aligned_offset_malloc (
	size_t	size,
	size_t	alignment,
	size_t	mod)
{
#ifdef _WIN64
	return (_aligned_offset_malloc (size, alignment, mod));
#else
	char	*p, *q;
	p = (char *) malloc (sizeof (void *) + size + alignment);
	if (p == NULL) return (NULL);
	q = (char *) (((long) p + sizeof (void *) + mod + alignment - 1) & ~(alignment - 1)) - mod;
	* (void **) ((char *) q - sizeof (void *)) = p;
	return (q);
#endif
}

void * aligned_malloc (
	size_t	size,
	size_t	alignment)
{
#ifdef _WIN64
	return (_aligned_malloc (size, alignment));
#else
	return (aligned_offset_malloc (size, alignment, 0));
#endif
}

void aligned_free (
	void	*ptr)
{
#ifdef _WIN64
	_aligned_free (ptr);
#else
	if (ptr == NULL) return;
	free (* (void **) ((char *) ptr - sizeof (void *)));
#endif
}
